""" Data parameters and algorithms. """
from dataclasses import dataclass
import numpy as np
import tensorflow as tf
from typing import Callable, Iterator, Optional, Tuple, Union

import masterful.enums
"""
A `DatasetLike` object conforms to a subset of inputs
that are supported by `tf.keras.Model.fit()`. These include:

- A tf.data.Dataset
- A tuple of numpy arrays, (x, y)
- A numpy array, x
- A Keras Sequence
- A generator and output signature. The generator is a function
  which returns an object which conforms to the `iter()` protocol.
"""
DatasetLike = Union[tf.data.Dataset, np.ndarray, Tuple[np.ndarray, np.ndarray],
                    tf.keras.utils.Sequence, Tuple[Callable[[], Iterator],
                                                   Tuple[tf.TensorSpec,
                                                         tf.TensorSpec]]]


@dataclass
class DataParams:
  """Parameters describing the datasets used during training.

  These parameters describe both the structure of the dataset
  (image and label shapes for examples) as well as semantic
  structure of the labels (the bounding box format for example,
  or whether or not the labels are sparse or dense).

  Args:
    num_classes: The number of possible classes in the dataset.
    task: The task this dataset will be used for.
    image_shape: The input shape of image data in the dataset,
      in the format (height, width, channels) if `input_channels_last=True`,
      otherwise (channels, height, width) if `input_channels_last=False`.
    image_range: The range of pixels in the input image space that
      of the dataset.
    image_dtype: The image data type in the dataset.
    image_channels_last: The ordering of the dimensions in the inputs.
      `input_channels_last=True` corresponds to inputs with shape
      (height, width, channels) while `input_channels_last=False`
      corresponds to inputs with shape (channels, height, width). Defaults
      to True.
    label_dtype: The data type of the labels.
    label_shape: The shape of the labels.
    label_structure: The tensor format of the label examples.
    label_sparse: True if the labels are in sparse format, False
      for dense (one-hot) labels.
    label_bounding_box_format: The format of bounding boxes in the label,
      if they exist.
  """
  num_classes: int = None
  task: "masterful.enums.Task" = None

  image_shape: Tuple = None
  image_range: "masterful.enums.ImageRange" = None
  image_dtype: tf.dtypes.DType = None
  image_channels_last: bool = True

  label_dtype: type = None
  label_shape: Tuple = None
  label_structure: "masterful.enums.TensorStructure" = None
  label_sparse: bool = None
  label_bounding_box_format: Optional[
      "masterful.enums.BoundingBoxFormat"] = None

  @property
  def image_width(self) -> int:
    """Gets the width of the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[1]
              if self.image_channels_last else self.image_shape[2])
    else:
      return self.image_shape[1]

  @property
  def image_height(self) -> int:
    """Gets the height of the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[0]
              if self.image_channels_last else self.image_shape[1])
    else:
      return self.image_shape[0]

  @property
  def image_channels(self) -> int:
    """Gets the number of channels in the images in this dataset."""
    if len(self.image_shape) > 2:
      return (self.image_shape[2]
              if self.image_channels_last else self.image_shape[0])
    else:
      return 0



"""Convenience functions."""
import tensorflow as tf
import typing


def learn_architecture_and_data_params(
    model: tf.keras.Model,
    datasets: typing.Sequence["masterful.data.DatasetLike"],
    task: "masterful.enums.Task",
    image_range: "masterful.enums.ImageRange",
    num_classes: int,
    prediction_logits: bool,
    backbone_only: bool = False,
    ensemble_multiplier: int = 1,
    sparse_labels: typing.Sequence[typing.Optional[bool]] = [],
    bounding_box_format: "masterful.enums.BoundingBoxFormat" = None,
) -> typing.Tuple["masterful.architecture.ArchitectureParams",
                  typing.Sequence["masterful.data.DataParams"]]:
  """Convenience function to jointly learn Architecture and Data Params.

  This function can help reduce the boilerplate necessary
  to learn the data and architecture parameters for your
  dataset(s) and models. In particular, you can jointly learn
  the data parameters for your training, test, unlabeled, and
  validation datasets in one function call. Importantly, they must
  all have the same image and label formats, and the
  model must also accept the same image format. If these
  conditions do not hold, then use the individual learners
  to find the correct parameters.

  Example:

  .. code-block:: python

    model: tf.keras.Model = ...
    training_dataset: tf.data.Dataset = ...
    unlabeled_dataset: td.data.Dataset = ...

    model_params, (training_dataset_params, unlabeled_dataset_params = masterful.training.learn_architecture_and_data_params(
        model=model,
        datasets=[training_dataset, unlabeled_dataset],
        task=masterful.enums.Task.CLASSIFICATION,
        image_range=masterful.enums.ImageRange.ZERO_255,
        num_classes=10,
        prediction_logits=True,
        sparse_labels=[False, None])

  Args:
    model: The model to learn the parameters for.
    datasets: A `tf.data.Dataset` instance to learn the parameters for.
    task: A semantic description of the a model's predictions or groundtruth
      labels, such as single-label classification (CIFAR10/Imagenet) and
      detection (COCO).
    image_range: The range of pixels in the input image space that
      the model is expecting.
    num_classes: The number of possible classes in the model prediction.
      If the model is a backbone only, then this should still be provided
      and represents the number of classes that should exist in in the
      classification head of the model.
    prediction_logits: True the model's output prediction is has no activation
      function. False means the model output is activated with a sigmoid
      or softmax layer.
    custom_objects: Some TF/Keras models have custom objects (layers,
      initializers, etc). If so, in order to load/save/clone the model, you
      need to let TF/Keras know how to load/save them. A dictionary of class
      mappings is the typical way to do this.
    backbone_only: True if this model represents a backbone (feature extractor)
      only, without a classification or task specific head. Defaults to False.
    ensemble_multiplier: The number of models to train as part of an ensemble.
      Defaults to 1, which disables ensembling and trains only a single model.
    sparse_labels: True if the labels are in sparse format, False
      for dense (one-hot) labels. Sequence of the same length as `datasets`.
    bounding_box_format: The format of bounding boxes in the label,
      if they exist.

  Returns:
    Tuple containing the model and a list containing the data parameters.
  """
  import masterful.architecture
  import masterful.data
  architecture_params = masterful.architecture.learn_architecture_params(
      model=model, task=task, input_range=image_range, num_classes=num_classes, prediction_logits=prediction_logits,
      backbone_only=backbone_only, ensemble_multiplier=ensemble_multiplier)

  # Learn all of the dataset parameters
  dataset_params = []
  if len(datasets) != len(sparse_labels):
    raise ValueError(
        "Arguments datasets and sparse_labels must have the same number of elements."
    )
  for i, dataset in enumerate(datasets):
    params = masterful.data.learn_data_params(dataset=dataset, image_range=image_range, task=task,
                                              num_classes=num_classes, sparse_labels=sparse_labels[i],
                                              bounding_box_format=bounding_box_format,)
    dataset_params.append(params)
  return architecture_params, dataset_params


"""Entrypoint for the masterful.train package."""
import os
# Must be set before importing tensorflow.
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['TF_CPP_MAX_LOG_LEVEL'] = '0'
os.environ['TF_CPP_MAX_VLOG_LEVEL'] = '0'

import argparse
import sys
import tensorflow as tf

gpus = tf.config.experimental.list_physical_devices('GPU')
for gpu in gpus:
  tf.config.experimental.set_memory_growth(gpu, True)

# Register if not running locally. Registration
# only works by using the evironment variables.
import masterful
if 'backend' not in dir(masterful) or os.environ.get('NOOP_REGISTER',
                                                     None) == '1':
  masterful = masterful.activate()
import masterful.train.trainer as trainer


def _silence_logging():
  tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

  import logging

  logging.getLogger('tensorflow').setLevel(logging.ERROR)
  logging.getLogger('absl').setLevel(logging.ERROR)
  logging.getLogger().setLevel(logging.ERROR)

  from absl import logging as absl_logging
  absl_logging.set_verbosity(absl_logging.ERROR)


_silence_logging()


def main(override=None):
  """
  Main entrypoint for the standalone version of this package.
  """
  # This script only runs under Tensorflow 2.0
  assert tf.__version__.startswith("2")

  parser = argparse.ArgumentParser(prog="masterful-train")
  parser.add_argument("--config",
                      help="Training configuration YAML file",
                      type=str,
                      required=True)
  parser.add_argument(
      "--output_path",
      help="Overrides the output path in the configuration YAML",
      type=str)
  parser.add_argument("--cache_path",
                      help="Overrides the cache path in the configuration YAML",
                      type=str)
  parser.add_argument(
      "--batch_size",
      help="Overrides the meta-learned batch size for finer control",
      type=int,
      default=-1)
  parser.add_argument(
      "--regularization_params",
      help=
      "[EXPERIMENTAL] Overrides the regularization parameters for training.",
      type=str)
  args = parser.parse_args()

  import warnings
  with warnings.catch_warnings():
    if masterful.constants.MASTERFUL_VERBOSITY < 1:
      warnings.simplefilter("ignore")
    trainer.train_from_config(
        config_path=args.config,
        output_path=args.output_path,
        cache_path=args.cache_path,
        regularization_params_path=args.regularization_params,
        batch_size=args.batch_size,
    )
  sys.exit(0)


main()


"""Entrypoint for the masterful.train package."""
import os
# Must be set before importing tensorflow.
os.environ['TF_CPP_MIN_LOG_LEVEL'] = '3'
os.environ['TF_CPP_MAX_LOG_LEVEL'] = '0'
os.environ['TF_CPP_MAX_VLOG_LEVEL'] = '0'

import argparse
import tensorflow as tf

# Register if not running locally. Registration
# only works by using the evironment variables.
import masterful
if 'backend' not in dir(masterful) or os.environ.get('NOOP_REGISTER',
                                                     None) == '1':
  masterful = masterful.register()
import masterful.data.duplicate_detector as duplicate_detector


def _silence_logging():
  tf.compat.v1.logging.set_verbosity(tf.compat.v1.logging.ERROR)

  import logging

  logging.getLogger('tensorflow').setLevel(logging.ERROR)
  logging.getLogger('absl').setLevel(logging.ERROR)
  logging.getLogger().setLevel(logging.ERROR)

  from absl import logging as absl_logging
  absl_logging.set_verbosity(logging.ERROR)


_silence_logging()


def main(override=None):
  """
  Main entrypoint for the standalone version of this package.
  """
  # This script only runs under Tensorflow 2.0
  assert tf.__version__.startswith("2")

  parser = argparse.ArgumentParser()
  parser.add_argument("--config",
                      help="Training configuration YAML file",
                      type=str,
                      required=True)
  args = parser.parse_args()

  duplicate_detector.find_duplicates(config_path=args.config)


main()


""" Data parameters and algorithms. """
from typing import Sequence

import masterful.enums
from masterful.data import DataParams


def learn_data_params(
    dataset: "masterful.data.DatasetLike",
    image_range: "masterful.enums.ImageRange",
    num_classes: int,
    sparse_labels: bool,
    task: "masterful.enums.Task",
    bounding_box_format: "masterful.enums.BoundingBoxFormat" = None,
) -> DataParams:
  """Learns the :class:`DataParams` for the given dataset.

  Most parameters can be introspected from the dataset itself.
  Anything that cannot be introspected is passed into this function
  as an argument, or set on the :class:`DataParams` after creation.

  Example:

  .. code-block:: python

    training_dataset: tf.data.Dataset = ...
    dataset_params = masterful.data.learn_data_params(
        dataset=training_dataset,
        image_range=masterful.enums.ImageRange.ZERO_255,
        num_classes=10,
        sparse_labels=False,
        task=masterful.enums.Task.CLASSIFICATION)

  Args:
    dataset: A `tf.data.Dataset` instance to learn the parameters for.
    image_range: The range of pixels in the input image space that
      of the dataset.
    num_classes: The number of possible classes in the dataset.
    sparse_labels: True if the labels are in sparse format, False
      for dense (one-hot) labels.
    task: The task this dataset will be used for.
    bounding_box_format: The format of bounding boxes in the label,
      if they exist.

  Returns:
    A new instance of DataParams describing the passed in dataset.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )


def learn_data_params_for_datasets(
    datasets: Sequence["masterful.data.DatasetLike"],
    image_range: "masterful.enums.ImageRange",
    num_classes: int,
    sparse_labels: Sequence[bool],
    task: "masterful.enums.Task",
    bounding_box_format: "masterful.enums.BoundingBoxFormat" = None,
) -> Sequence[DataParams]:
  """Learns the :class:`DataParams` for the given datasets.

  Convenience method for learning the data parameters for multiple
  datasets at a time.

  Example:

  .. code-block:: python

    # Learn parameters for three datasets at the same time
    training_dataset: tf.data.Dataset = ...
    validation_dataset: tf.data.Dataset = ...
    test_dataset: tf.data.Dataset = ...

    (training_dataset_params, validation_dataset_params, test_dataset_params) = masterful.data.learn_data_params(
        datasets=[training_dataset, validation_dataset, test_dataset),
        image_range=masterful.enums.ImageRange.ZERO_255,
        num_classes=10,
        sparse_labels=[False, False, False],
        task=masterful.enums.Task.CLASSIFICATION)


  Args:
    datasets: A list/tuple of `tf.data.Dataset` instance to learn the parameters for.
    image_range: The range of pixels in the input image space that
      of the dataset.
    num_classes: The number of possible classes in the dataset.
    sparse_labels: A list/tuple of True if the labels are in sparse format, False
      for dense (one-hot) labels.
    task: The task this dataset will be used for.
    bounding_box_format: The format of bounding boxes in the label,
      if they exist.

  Returns:
    A sequence of DataParams instances describing the passed in datasets.
  """
  raise RuntimeError(
      "Please call masterful.register() with your account ID and authorization key before using the API."
  )
